#include "contour_creator.hh"
#include <gdal/ogr_api.h>
#include "ogrsf_frmts.h"
#include <gdal/ogr_core.h>
#include <gdal/ogr_feature.h>
#include <gdal/ogr_geometry.h>
#include <iostream>
#include <iterator>
#include <ostream>
#include <vector>
#include <cstdint>
#include <gdal/gdal.h>
#include "gdal/gdal_priv.h"
#include <gdal/gdal_frmts.h>
#include <omp.h>

std::vector<Point> produce_cellmap(HeightMap* heightmap, float z)
{
    int length = (heightmap->width-1)*(heightmap->height-1); // Defining length of vector
    uint8_t *cells = (uint8_t *) CPLMalloc(sizeof(uint8_t)*length);
    std::vector<Point> points; // Initiating a vector of points
    for (int i = 0; i<length; i++) {
        int y = i/(heightmap->width-1);
        int x = i%(heightmap->width-1);
        uint8_t result = (heightmap->get_pixel(x,y)>z) + 
        (heightmap->get_pixel(x+1,y)>z)*2 +
        (heightmap->get_pixel(x+1,y+1)>z)*4 +
        (heightmap->get_pixel(x,y+1)>z)*8;
        if (result != 0 and result != 15 ) {
            points.push_back(Point(x, y, result));

        };
    }
    
    return points;


}
std::tuple<double, double> local_to_projected(double* geotransform, int x, int y)
{
    return {
        geotransform[1] * x + geotransform[2] * y +
        geotransform[1] * 0.5 + geotransform[2] * 0.5 + geotransform[0],
        geotransform[4] * x + geotransform[5] * y +
        geotransform[4] * 0.5 + geotransform[5] * 0.5 + geotransform[3]
        };
}
std::vector<OGRFeature*> create_contours(OGRFeatureDefn* layer_def, std::vector<Point> points, HeightMap* heightmap)
{
    std::vector<OGRFeature*> lines;
    for (int i = 0; i<points.size();i++)
    {
        OGRFeature *feature;
            OGRLineString *geometry = new OGRLineString();
            
            feature = OGRFeature::CreateFeature(layer_def);
            feature->SetField( "Name", i );
        Point* current = &points[0];
        int points_delegated = 0;
        while (points_delegated < points.size()) 
        {
            //int x_raw = i%width;
            //int y_raw = i/height;
            int x_raw = current->x;
            int y_raw = current->y;
            auto [x, y] = local_to_projected(heightmap->geotransform, x_raw, y_raw);
            //std::cout << x << ", " << y << "  ";
            //auto [x_bl, y_bl, x_tr, y_tr] = marching_squares_lookup(*(cellmap->cells + i) );
            //std::cout << x << ", " << y << "  ";
            //geometry->setPoint( geometry->getNumPoints(), x+x_tr*0.25, y+y_tr*0.25 ); 
            //geometry->setPoint( geometry->getNumPoints(), x+x_bl*0.25, y+y_bl*0.25 ); 
            geometry->setPoint(geometry->getNumPoints(),x ,y );
            points_delegated++;
            current = current->next;
            if (current == nullptr) 
            {
                for (int k = 0; k < points.size(); k++)
                {
                    if (points[k].previous == nullptr)
                    {
                    current = &points[k];
                    }
                }
                if (current == nullptr)
                {
                    break;
                }

            }
        }

                    if ( feature->SetGeometry(geometry) != OGRERR_NONE)
            {
                printf( "Failed to set geometry.\n" );
                exit( 1 );
            }
            OGRGeometryFactory::destroyGeometry(geometry);
            lines.push_back(feature);
    }
    
}

std::vector<OGRFeature*> vector_cellmap(OGRFeatureDefn* layer_def, HeightMap* heightmap, int interval)
{
    int num_contours = (heightmap->max - heightmap->min)/interval;
    std::vector<OGRFeature*> vector_contours;
    omp_set_num_threads(12);

    #pragma omp parallel
    {
        std::vector<OGRFeature*> lines;
        #pragma omp for
        for (int i = 1; i <= num_contours; i++) 
        {
            auto points = produce_cellmap(heightmap, heightmap->min + interval*i);

                
            for (int j = 0; j < points.size(); j++){
                Point* current = &points[j];
                //std::cout << points_allocated << " " << points.size() << "\n";
                for (int k = 0; k < points.size(); k++){
                    Point* candidate = &points[k];
                    if (candidate->previous == nullptr) {
                        
                        if (candidate->x +1 == current->x && candidate->y == current->y) {
                            current->next = candidate;
                            candidate->previous = current;
                        }
                        else if (candidate->x -1 == current->x && candidate->y == current->y) {
                            current->next = candidate;
                            candidate->previous = current;
                        }
                        else if (candidate->y +1 == current->y && candidate->x == current->x) {
                            current->next = candidate;
                            candidate->previous = current;
                        }
                        else if (candidate->y -1 == current->y && candidate->x == current->x) {
                            current->next = candidate;
                            candidate->previous = current;
                        }
                        
                    }
                }
            }
            auto lines = create_lines(layer_def, points, heightmap);
            
        }
        
        #pragma omp critical

        vector_contours.insert(vector_contours.end(), lines.begin(), lines.end());

    }
        return vector_contours;
}



constexpr std::tuple<double, double, double, double> marching_squares_lookup(int mcase)
{
    switch (mcase) {
        case 1: return {0, 0.5, 0.5, 0};
        case 2: return {1, 0.5, 0.5, 0};
        case 3: return {0, 0.5, 1, 0};
        case 4: return {0.5, 1, 1, 0.5};
        case 5: return {0, 0, 0, 0}; //FIXME
        case 6: return {0.5, 1, 0.5, 0};
        case 7: return {0, 0.5, 0.5, 1};
        case 8: return {0, 0.5, 0.5, 1};
        case 9: return {0.5, 1, 0.5, 0};
        case 10: return {0, 0, 0, 0}; //FIXME
        case 11: return {0.5, 1, 1, 0.5};
        case 12: return {0, 0.5, 1, 0.5};
        case 13: return {0.5, 0, 1, 0.5};
        case 14: return {0, 0.5, 0.5, 0};
    };
}


void write_output_file(const char *filepath, HeightMap* heightmap)
{

    const char *pszDriverName = "GeoJSON";
    GDALDriver *poDriver;

    GDALAllRegister();

    poDriver = GetGDALDriverManager()->GetDriverByName(pszDriverName );
    if( poDriver == NULL )
    {
        printf( "%s driver not available.\n", pszDriverName );
        exit( 1 );
    }

    GDALDataset *poDS;

    poDS = poDriver->Create( filepath, 0, 0, 0, GDT_Unknown, NULL );
    if( poDS == NULL )
    {
        printf( "Creation of output file failed.\n" );
        exit( 1 );
    }

    OGRLayer *poLayer;

    poLayer = poDS->CreateLayer( "contours", /*&(cellmaps[0]).reference_system*/ NULL, wkbLineString, NULL );
    if( poLayer == NULL )
    {
        printf( "Layer creation failed.\n" );
        exit( 1 );
    }

    OGRFieldDefn oField( "Name", OFTString );

    oField.SetWidth(32);

    if( poLayer->CreateField( &oField ) != OGRERR_NONE )
    {
        printf( "Creating Name field failed.\n" );
        exit( 1 );
    }
    int width = heightmap->width -1;
    int height = heightmap->height -1;
    int size = width * height;

    // gets the lines from the heightmap. Runs the actual algorithm
    auto lines = create_contours(poLayer->GetLayerDefn(), heightmap, 5);


    for (int j = 0; j < lines.size(); j++)
    {

            if( poLayer->CreateFeature( lines[j] ) != OGRERR_NONE )
            {
                printf( "Failed to create feature in shapefile.\n" );
                exit( 1 );
            }
            OGRFeature::DestroyFeature( lines[j] );  
        

    }

    GDALClose( poDS );
}

int main(int argc, const char* argv[])
{
    const char* filepath = argv[1];
    HeightMap map(filepath); 
    
    std::cout << "x: " << map.width << "    y: " << map.height << "\n";
    std::cout << "max: " << map.max << "    min: " << map.min << "\n";

    map.blur(0.8);
    
    auto cellmap = produce_cellmap(&map, 40);
    
    
    write_output_file("out.geojson", &map);
}